# Listeners for the default models, which learn forward and reverse word chains
# from a catchAll block and respond to "hubot markov" and "hubot remarkov".
# If configured to do so, randomly respond to arbitrary messages with a
# markov string generated from a word from that message.

processors = require './processors'

module.exports = (robot, config) ->
  activeModelNames = []

  reportErr = (msg, err) ->
    msg.send ":boom:\n```#{err.stack}```"

  if config.defaultModel
    robot.markov.createModel 'default_forward', {}
    activeModelNames.push 'default_forward'

    robot.markov.generateForward = (seed, callback) ->
      robot.markov.modelNamed 'default_forward', (model) ->
        model.generate seed, config.generateMax, callback

    # Generate markov chains on demand, optionally seeded by some initial state.
    robot.respond /markov(\s+(.+))?$/i, (msg) ->
      robot.markov.generateForward msg.match[2] or '', (err, text) ->
        return reportErr(err) if err?
        msg.send text

  if config.reverseModel
    robot.markov.createModel 'default_reverse', {}, (model) ->
      model.processWith processors.reverseWords

    activeModelNames.push 'default_reverse'

    robot.markov.generateReverse = (seed, callback) ->
      robot.markov.modelNamed 'default_reverse', (model) ->
        model.generate seed, config.generateMax, callback

    # Generate reverse markov chains on demand, optionally seeded by some end state
    robot.respond /remarkov(\s+(.+))?$/i, (msg) ->
      robot.markov.generateReverse msg.match[2] or '', (err, text) ->
        return reportErr(err) if err?
        msg.send text

  if config.defaultModel and config.reverseModel

    robot.markov.generateMiddle = (seed, callback) ->
      generateRight = (cb) ->
        robot.markov.modelNamed 'default_forward', (model) ->
          model.generate seed, config.generateMax, cb

      generateRest = (right, cb) ->
        words = processors.words.pre(right)
        rightSeed = words.shift() or ''

        robot.markov.modelNamed 'default_reverse', (model) ->
          model.generate rightSeed, config.generateMax, (err, left) ->
            return cb(err) if err?
            cb(null, [left, words...].join ' ')

      generateRight (err, right) ->
        return callback(err) if err?
        generateRest right, callback

    # Generate markov chains with the seed in the middle
    robot.respond /mmarkov(\s+(.+))?$/i, (msg) ->
      robot.markov.generateMiddle msg.match[2] or '', (err, text) ->
        return reportErr(err) if err?
        msg.send text

  if activeModelNames.length isnt 0

    learningListener = (msg) ->
      # Ignore empty messages
      return if !msg.message.text

      # Return if message containers a URL
      return if !config.includeUrls and msg.message.text.match /https?:\/\//

      # Disregard ignored usernames.
      return if msg.message.user.name in config.ignoreList

      # Disregard any messages that have keywords
      for phrase in config.ignoreMessageList
        return if msg.message.text.indexOf(phrase) isnt -1

      # Pass the message to each active model.
      for name in activeModelNames
        robot.markov.modelNamed name, (model) -> model.learn msg.message.text

    if config.learningListenMode == 'hear-all'
      robot.hear /.*/i, learningListener
    else if config.learningListenMode == 'catch-all'
      robot.catchAll learningListener
    else
      robot.hear RegExp("" + config.learningListenMode), learningListener

    if config.respondChance > 0
      respondListener = (msg) ->
        if Math.random() < config.respondChance
          randomWord = msg.random(processors.words.pre(msg.message.text)) or ''

          if config.reverseModel
            robot.markov.generateMiddle randomWord, (text) -> msg.send text
          else
            robot.markov.generateForward randomWord, (text) -> msg.send text

      if config.respondListenMode == 'hear-all'
           robot.hear /.*/i, respondListener
         else if config.respondListenMode == 'catch-all'
           robot.catchAll respondListener
         else
           robot.hear RegExp("" + config.respondListenMode), respondListener
